/*
 * Copy a file into a filesystem.
 *
 * Copyright (C) 2019, 2022, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <string.h>

#include <sys/stat.h>

#include <e2access/path.h>
#include <fsclient/client.h>
#include <systypes/fcntl.h>

#include "file.h"
#include "ops.h"
#include "session.h"



/* Copy buffer size. */

static int BUFSIZE = 4096;



/* Alternative metadata set by options. */

extern struct metadata md;



/* Copy a file into the filesystem image. */

static int _copy_file_in(const char *source, const char *target, flags_t flags)
{
    int retval = 0;

    /* Copying details. */

    FILE *fp;
    file_t *target_fp;
    char buf[BUFSIZE];
    size_t got;
    offset_t written;

    /* Open a file in the target directory. */

    target_fp = client_open(target, flags);

    if (target_fp == NULL)
        return 1;

    /* Open the file in the source directory. */

    fp = fopen(source, "r");

    /* Copy the file content. */

    if (fp != NULL)
    {
        while ((got = fread(buf, sizeof(char), BUFSIZE, fp)))
        {
            while (got)
            {
                written = client_write(target_fp, buf, got);

                if (!written)
                {
                    retval = 1;
                    goto close_files;
                }
                got -= written;
            }
        }
    }

close_files:
    fclose(fp);
    client_close(target_fp);

    return retval;
}

/* Copy a source file from the external environment into the filesystem. */

static int _copy_in(const char *source, const char *target,
                    const char *basename)
{
    /* Obtain basename and eventual target path. */

    const char *source_basename = path_basename(source);
    char target_plus_basename[strlen(target) + 1 +
                              (basename != NULL ? strlen(basename)
                                                : strlen(source_basename)) + 1];
    const char *target_path;

    /* Target file properties. */

    flags_t flags;
    struct stat st;
    long err;

    /* By default, treat the target as a new object. */

    int target_is_new = 1;

    /* Without a basename, the target exists and is either a directory, into
       which the source file shall be copied, or it is a file that shall be
       overwritten. */

    if (basename == NULL)
    {
        err = client_stat(target, &st);

        if (err)
        {
            fprintf(stderr, "Could not stat target: %s\n", target);
            return 1;
        }

        target_is_new = S_ISDIR(st.st_mode);

        if (target_is_new)
            basename = path_basename(source);
    }

    /* Obtain the target path. */

    if (basename != NULL)
    {
        /* At the top level, omit the directory. */

        if (strlen(target))
        {
            sprintf(target_plus_basename, "%s/%s", target, basename);
            target_path = target_plus_basename;
        }
        else
            target_path = basename;
    }
    else
        target_path = target;

    /* Directories are created with the same metadata. */

    if (isdir(source))
    {
        if (!target_is_new)
        {
            fprintf(stderr, "Directory cannot be copied as it already exists: %s\n", target_path);
            return 1;
        }

        err = client_stat(source, &st);

        if (err)
        {
            fprintf(stderr, "Could not stat source: %s\n", source);
            return 1;
        }

        if (client_mkdir(target_path, st.st_mode & ~md.mask))
            return 1;
    }

    /* Files are copied. */

    else if (isfile(source))
    {
        flags = O_WRONLY;

        /* Obtain the inode for the target file. */

        if (target_is_new)
            flags |= O_CREAT;

        /* NOTE: Overwrite/update metadata where appropriate. */

        if (_copy_file_in(source, target_path, flags))
        {
            fprintf(stderr, "Failed to copy file: %s\n", source);
            return 1;
        }
    }

    return 0;
}

/* Copy source files from the external environment into the filesystem image. */

int copy_in(int argc, char *argv[])
{
    int i;

    /* Target filename details. */

    char *target = argv[argc - 1];
    char *basename;
    struct stat st;

    /* Locate the target and test whether it is a file or a directory. */

    long err = client_stat(target, &st);

    /* Only a non-existent file in an existing directory is permitted. */

    if (err == -L4_ENOENT)
    {
        /* Split the path, making target the parent directory. */

        basename = path_split(target);

        /* Set the parent directory to the empty string if absent. */

        if (basename == target)
            target = basename + strlen(basename);

        err = client_stat(target, &st);

        if (err)
        {
            fprintf(stderr, "Could not stat target parent: %s\n", target);
            return 1;
        }

        if (!S_ISDIR(st.st_mode))
        {
            fprintf(stderr, "Target parent is not directory: %s\n", target);
            return 1;
        }
    }
    else if (err)
    {
        fprintf(stderr, "Could not stat target: %s\n", target);
        return 1;
    }

    /* An existing object can be a file or directory. */

    else
    {
        basename = NULL;

        /* Only permit a target file when one source file is given. */

        if (!S_ISDIR(st.st_mode) && (argc > 2))
        {
            fprintf(stderr, "Target can only be a file when copying a single file: %s\n", target);
            return 1;
        }
    }

    /* Copy each source object to the target directory. */

    for (i = 0; i < argc - 1; i++)
        if (_copy_in(argv[i], target, basename))
            return 1;

    return 0;
}

/* vim: tabstop=4 expandtab shiftwidth=4
*/
